#![allow(dead_code)]

mod util;

pub fn init() -> Interface { Interface::new() }

type Class = u8;
type Handle = u64;

const TAG_UNDEFINED :Class = 0x00;
const TAG_NULL      :Class = 0x01;
const TAG_BOOLEAN   :Class = 0x02;
//...
const TAG_NATURAL   :Class = 0x10;
const TAG_INTEGER   :Class = 0x11;
const TAG_DECIMAL   :Class = 0x12;
//...
const TAG_BLOCK     :Class = 0x1e;
const TAG_STRING    :Class = 0x1f;
const TAG_OPTION    :Class = 0x20;
const TAG_SET       :Class = 0x21;
const TAG_ARRAY     :Class = 0x22;
const TAG_LIST      :Class = 0x23;
const TAG_SPARSE    :Class = 0x24;
const TAG_MAP       :Class = 0x25;
const TAG_TREE      :Class = 0x26;
const TAG_GRAPH     :Class = 0x27;
//...
const TAG_RECORD    :Class = 0x3e;
const TAG_SCHEMA    :Class = 0x3f;

fn mask_tag   (tag:Class)     -> Handle { (tag as u64) << 56 }
fn unmask_tag (handle:Handle) -> Class  { (handle >> 56) as Class }
fn handle_id  (handle:Handle) -> usize  { (handle & !mask_tag(0xff)) as usize }

struct Container<T> {
    pub class:Vec<Class>,
    pub data:T,
}

pub struct Interface {
    pool_boolean:stdu::Pool<bool>,
    pool_natural:stdu::Pool<u64>,
    pool_integer:stdu::Pool<i64>,
    pool_block:stdu::Pool<Vec<u8>>,
    pool_string:stdu::Pool<String>,
    pool_option:stdu::Pool<Container<Option<Handle>>>,
    pool_array:stdu::Pool<Container<Vec<Handle>>>,
    pool_list:stdu::Pool<Container<Vec<Handle>>>,
    pool_sparse:stdu::Pool<Container<stdu::Sparse<Handle>>>,
}
impl Interface {
    pub(crate) fn new() -> Self
    {
        Self {
            pool_boolean:stdu::Pool::<bool>::new(),
            pool_natural:stdu::Pool::<u64>::new(),
            pool_integer:stdu::Pool::<i64>::new(),
            pool_block:stdu::Pool::<Vec<u8>>::new(),
            pool_string:stdu::Pool::<String>::new(),
            pool_option:stdu::Pool::<Container<Option<Handle>>>::new(),
            pool_array:stdu::Pool::<Container<Vec<Handle>>>::new(),
            pool_list:stdu::Pool::<Container<Vec<Handle>>>::new(),
            pool_sparse:stdu::Pool::<Container<stdu::Sparse<Handle>>>::new(),
        }
    }


    /* DATA CONSTRUCTORS */
 
    pub fn boolean(&mut self, value:bool) -> Handle {
        mask_tag(TAG_BOOLEAN) + self.pool_boolean.add(value) as Handle
    }

    pub fn natural(&mut self, value:u64) -> Handle {
        mask_tag(TAG_NATURAL) + self.pool_natural.add(value) as Handle
    }

    pub fn integer(&mut self, value:i64) -> Handle {
        mask_tag(TAG_INTEGER) + self.pool_integer.add(value) as Handle
    }

    pub fn block(&mut self, value:&Vec<u8>) -> Handle {
        mask_tag(TAG_BLOCK) + self.pool_block.add(value.clone()) as Handle
    }

    pub fn string(&mut self, value:&str) -> Handle {
        mask_tag(TAG_STRING) + self.pool_string.add(value.to_string()) as Handle
    }


    /* ENCODING/DECODING */

    fn encode_tags(tags:Vec<Class>) -> Vec<u8>
    {
        let mut encoded = Vec::<u8>::with_capacity(16);
        for tag in tags {
            encoded.append(&mut util::pack_natural(tag as u64));
        }
        return encoded;
    }

    fn encode_parts(&self, handle:Handle) -> Result<(Vec<Class>, Vec<u8>), ()>
    {
        let mut tags = Vec::<Class>::with_capacity(3);
        let mut encoded = Vec::<u8>::with_capacity(64);

        match unmask_tag(handle) {
            TAG_NULL => {
                tags.push(TAG_NULL);
            }
            TAG_BOOLEAN => match self.pool_boolean.get(handle_id(handle)) {
                Some(value) => {
                    tags.push(TAG_BOOLEAN + *value as Class);

                }
                None => { return Err(()); }
            }
            TAG_NATURAL => match self.pool_natural.get(handle_id(handle)) {
                Some(value) => {
                    tags.push(TAG_NATURAL);
                    encoded.append(&mut util::pack_natural(*value));
                }
                None => { return Err(()); }
            }
            TAG_INTEGER => match self.pool_integer.get(handle_id(handle)) {
                Some(value) => {
                    tags.push(TAG_INTEGER);
                    encoded.append(&mut util::pack_integer(*value));
                }
                None => { return Err(()); }
            }
            TAG_BLOCK => match self.pool_block.get(handle_id(handle)) {
                Some(data) => {
                    tags.push(TAG_BLOCK);
                    for b in data {
                        encoded.push(*b);
                    }
                }
                None => { return Err(()); }
            }
            TAG_STRING => match self.pool_string.get(handle_id(handle)) {
                Some(data) => {
                    tags.push(TAG_STRING);
                    let mut bytes = Vec::<u8>::from(data.as_bytes());
                    encoded.append(&mut util::pack_natural(bytes.len() as u64));
                    encoded.append(&mut bytes);
                }
                None => { return Err(()); }
            }
            TAG_ARRAY => match self.pool_array.get(handle_id(handle)) {
                Some(object) => {
                    tags.push(TAG_ARRAY);
                    tags.append(&mut object.class.clone());

                }
                None => { return Err(()); }
            }
            TAG_LIST => match self.pool_list.get(handle_id(handle)) {
                Some(object) => {
                    tags.push(TAG_LIST);
                    tags.append(&mut object.class.clone());

                }
                None => { return Err(()); }
            }
            TAG_SPARSE => match self.pool_sparse.get(handle_id(handle)) {
                Some(object) => {
                    tags.push(TAG_SPARSE);
                    tags.append(&mut object.class.clone());

                    // lookup table


                    // data table

                }
                None => { return Err(()); }
            }
            _ => { return Err(()); }
        }
        Err(())
    }

    pub fn encode_data(&self, handle:Handle) -> Result<Vec<u8>,()>
    {
        match self.encode_parts(handle) {
            Ok((_, encoded)) => Ok(encoded),
            Err(_) => Err(())
        }
    }

    pub fn encode(&self, handle:Handle) -> Result<Vec<u8>,()>
    {
        match self.encode_parts(handle) {
            Ok((tags, data)) => {
                let prefix = Self::encode_tags(tags);
                let mut encoded = Vec::<u8>::with_capacity(prefix.len() + data.len());
                for b in prefix { encoded.push(b); }
                for b in data { encoded.push(b); }

                Ok(encoded)
            }
            Err(_) => Err(())
        }
    }

    fn decode_tags(data:&Vec<u8>, index:&mut usize) -> Vec<Class>
    {
        let mut tags = Vec::<Class>::new();
        while *index < data.len() {
            let tag = util::unpack_natural(data, index) as Class;
            tags.push(tag);
            match tag {
                TAG_ARRAY | TAG_LIST | TAG_SPARSE |TAG_RECORD => { }
                _ => { break; }
            }
        }
        return tags;
    }

    pub fn decode(&mut self, data:&Vec<u8>) -> Result<Handle,()>
    {
        let mut index :usize = 0;
        let tags = Self::decode_tags(data, &mut index);

        for tag in tags {
            println!("{}", tag);
        }

        Err(())
    }
}
