#![allow(dead_code)]

mod util;
mod template; pub use template::Type;
mod interface; pub use interface::*;

pub type Class = u32;
pub type Handle = usize;

#[derive(Clone)]
pub enum Szun {
    /*00*/ Null,
    /*02*/ Boolean(bool),
    /*10*/ Natural(u64),
    /*11*/ Integer(i64),
    /*12*/ //Decimal(f64),
    /*20*/ Block(Vec<u8>),
    /*21*/ String(std::string::String),
    /*30*/ //Set(Set),
    /*31*/ Array(ArrayData),
    /*32*/ List(ListData),
    /*33*/ Sparse(SparseData),
    /*34*/ //Map(Map),
    /*35*/ //Tree(Tree),
    /*36*/ //Graph(Graph),
    /*40*/ //Enum(Enum),
    /*41*/ //Selection(Selection),
    /*4e*/ //Record(Record),
    /*4f*/ //Schema(Schema),
}
impl Szun {
    pub fn tag(&self) -> Class
    {
        match self {
            Self::Null => Null::tag(),
            Self::Boolean(_) => Boolean::tag(),
            Self::Natural(_) => Natural::tag(),
            Self::Integer(_) => Integer::tag(),
            Self::Block(_) => Block::tag(),
            Self::String(_) => String::tag(),
            Self::Array(_) => Array::tag(),
            Self::List(_) => List::tag(),
            Self::Sparse(_) => Sparse::tag(),
        }
    }
}

macro_rules! szun_as {
    ($name:ident, $class:pat, $impl:ident) => {
        pub fn $name(&mut self, handle:usize) -> Option<$impl>
        {
            match self.data.get(handle) {
                Some(object) => {
                    match object.object {
                        $class => { return Some($impl { instance: self, target: handle }); }
                        _ => { }
                    }
                }
                None => { }
            }
            return None;
        }
    };
}

pub fn init() -> Instance { Instance::new() }

struct Entry {
    owner:Handle,
    class:Class,
    object:Szun,
}
impl Entry {
    pub fn new(object:Szun) -> Self
    {
        Self {
            owner:0,
            class:0,
            object:object,
        }
    }

    pub fn new_classed(class:Class, object:Szun) -> Self
    {
        Self {
            owner:0,
            class:class,
            object:object,
        }
    }
}

pub struct Instance {
    pub(crate) schemas:stdu::Sparse<Szun>,
    pub(crate) data:stdu::Pool<Entry>,
}
impl Instance {
    pub(crate) fn new() -> Self
    {
        Self {
            schemas:stdu::Sparse::<Szun>::new(),
            data:stdu::Pool::<Entry>::new(),
        }
    }

    pub fn null(&mut self) -> Handle { self.data.add(Entry::new(Szun::Null)) }
    pub fn boolean(&mut self, data:bool) -> Handle { self.data.add(Entry::new(Szun::Boolean(data))) }
    pub fn natural(&mut self, data:u64) -> Handle { self.data.add(Entry::new(Szun::Natural(data))) }
    pub fn integer(&mut self, data:i64) -> Handle { self.data.add(Entry::new(Szun::Integer(data))) }
    pub fn block(&mut self, data:Vec<u8>) -> Handle { self.data.add(Entry::new(Szun::Block(data))) }
    pub fn string(&mut self, data:&str) -> Handle { self.data.add(Entry::new(Szun::String(data.to_string()))) }
    //pub fn array(&mut self, data:Vec<Handle>) -> Handle { self.data.add(Szun::Array(data)) }
    pub fn list(&mut self, class:Class, data:&Vec<Handle>) -> Option<Handle> {
        for handle in data {
            match self.data.get(*handle) {
                Some(object) => {
                    if object.object.tag() != class && class != 0 { return None; }
                }
                None => { return None; }
            }
        }

        let obj = Szun::List(ListData { class:class, data:data.clone() });

        return Some(self.data.add(Entry::new(obj)));
    }
    //pub fn sparse(&mut self, data:stdu::Sparse<Handle>) -> Handle { self.data.add(Szun::Sparse(data)) }

    szun_as!(as_null, Szun::Null, Null);
    szun_as!(as_boolean, Szun::Boolean(_), Boolean);
    szun_as!(as_natural, Szun::Natural(_), Natural);
    szun_as!(as_integer, Szun::Integer(_), Integer);
    szun_as!(as_block, Szun::Block(_), Block);
    szun_as!(as_string, Szun::String(_), String);

    fn encode_tags(tags:Vec<Class>) -> Vec<u8>
    {
        let mut encoded = Vec::<u8>::with_capacity(16);
        for tag in tags {
            encoded.append(&mut util::pack_natural(tag as u64));
        }
        return encoded;
    }

    fn encode_parts(&self, object:&Szun) -> Result<(Vec<Class>, Vec<u8>), ()>
    {
        let mut tags = Vec::<Class>::with_capacity(1);
        let mut encoded = Vec::<u8>::with_capacity(64);

        match object {
            Szun::Null => {
                tags.push(Null::tag());
            }
            
            Szun::Boolean(data) => {
                tags.push(if *data { Boolean::tag() + 1 } else { Boolean::tag() });
            }
            
            Szun::Natural(data) => {
                tags.push(Natural::tag());
                encoded.append(&mut util::pack_natural(*data));
            }
            
            Szun::Integer(data) => {
                tags.push(Integer::tag());
                encoded.append(&mut &mut util::pack_integer(*data));
            }
            
            //Szun::Decimal(_data) => {
            //    tag = Decimal::tag();
            //}
            
            Szun::Block(data) => {
                tags.push(Block::tag());
                tags.push(data.len() as Class);
                for b in data {
                    encoded.push(*b);
                }
            }
            
            Szun::String(data) => {
                tags.push(String::tag());
                let mut bytes = Vec::<u8>::from(data.as_bytes());
                let mut size = util::pack_natural(bytes.len() as u64);
                encoded.append(&mut size);
                encoded.append(&mut bytes);
            }
            
            //Szun::Set(_data) => {
            //    tag = Set::tag();
            //}
            
            Szun::Array(data) => {
                tags.push(Array::tag());
                tags.push(data.data.len() as Class);
                tags.push(data.class);

                for item in &data.data {
                    match self.data.get(*item) {
                        Some(entry) => {
                            match self.encode_parts(&entry.object) {
                                Ok((s_tags, mut s_data)) => {
                                    if data.class == 0 {
                                        encoded.append(&mut Self::encode_tags(s_tags));
                                    }
                                    encoded.append(&mut s_data);
                                }
                                Err(_) => { return Err(()); }
                            }
                        }
                        None => { return Err(()); }
                    }
                }
            }
            
            Szun::List(data) => {
                tags.push(List::tag());
                tags.push(data.class);

                encoded.append(&mut util::pack_natural(data.data.len() as u64));
                for item in &data.data {
                    match self.data.get(*item) {
                        Some(entry) => {
                            match self.encode_parts(&entry.object) {
                                Ok((s_tags, mut s_data)) => {
                                    if data.class == 0 {
                                        encoded.append(&mut Self::encode_tags(s_tags));
                                    }
                                    encoded.append(&mut s_data);
                                }
                                Err(_) => { return Err(()); }
                            }
                        }
                        None => { return Err(()); }
                    }
                }
            }
            
            Szun::Sparse(data) => {
                tags.push(Sparse::tag());
                tags.push(data.class);

                encoded.append(&mut util::pack_natural(data.data.len() as u64));
                //for item in &data {
                //    let (s_tags, mut s_data) = item.encode_parts();
                //    if data.class == 0 {
                //        encoded.append(&mut Szun::encode_tags(s_tags));
                //    }
                //    encoded.append(&mut s_data);
                //}
            }
            
            /*Szun::Map(_data) => {
                tag = Map::tag();
            }
            
            Szun::Enum(_data) => {
                tag = Enum::tag();
            }
            
            Szun::Selection(_data) => {
                tag = Selection::tag();
            }
            
            Szun::Record(_data) => {
                tag = Record::tag();
            }

            Szun::Schema(_data) => {
                tag = Schema::tag();
            }*/
        }

        return Ok((tags, encoded));
    }

    pub fn encode_data(&self, handle:Handle) -> Result<Vec<u8>,()>
    {
        return match self.data.get(handle) {
            Some(entry) => {
                match self.encode_parts(&entry.object) {
                    Ok((_, encoded)) => Ok(encoded),
                    Err(_) => Err(())
                }
            }
            None => Err(())
        };
    }

    pub fn encode(&self, handle:Handle) -> Result<Vec<u8>,()>
    {
        return match self.data.get(handle) {
            Some(entry) => {
                match self.encode_parts(&entry.object) {
                    Ok((tags, data)) => {
                        let prefix = Self::encode_tags(tags);
                        let mut encoded = Vec::<u8>::with_capacity(prefix.len() + data.len());
                        for b in prefix { encoded.push(b); }
                        for b in data { encoded.push(b); }

                        Ok(encoded)
                    },
                    Err(_) => Err(())
                }
            }
            None => Err(())
        };
    }
}
